// Copyright (c) 2017, Herman Bergwerf. All rights reserved.
// Use of this source code is governed by a MIT-style license
// that can be found in the LICENSE file.

import 'dart:html';
import 'dart:async';

import 'package:katex_js/katex_js.dart' as katex;

const _cursorCommand = r'\phantom{~}';
const _emptyCommand = r'\phantom{1}';

class EdiTeXCommand {
  final String name, latexTemplate, parseTemplate;

  /// Indicates if the height for super or subscript operators directly
  /// following the command is different. This way measures can be taken to
  /// prevent the cursor from disturbing the rendering.
  final bool changesHeight;

  const EdiTeXCommand(this.name, this.latexTemplate, this.parseTemplate,
      [this.changesHeight = false]);
}

/// Interface between editor and environment. When the user enters a command
/// this function is called to find a command template. If null is returned, the
/// command is considered unknown.
typedef EdiTeXCommand CommandResolver(String command);

/// Same as [CommandResolver] but for instant commands.
typedef EdiTeXCommand InstantCommandResolver(String command);

/// Command used when the user types '/'. This cannot be configured. It is
/// considered a good practice to stick to LaTeX like commands when those are
/// present.
const _divideCommand = 'frac';

const _defaultCommands = const {
  'frac': //
      const EdiTeXCommand('frac', r'\frac{~$0~}{~$1~}', r'frac($0,$1)', true),
  'sqrt': //
      const EdiTeXCommand('sqrt', r'\sqrt{$0}~', r'sqrt($0)'),
  'root': //
      const EdiTeXCommand('root', r'\sqrt[$0]{$1}~', r'root($0,$1)'),
  'log': //
      const EdiTeXCommand('log', r'\log_{$0}$1~', r'log($0,$1)'),
  'diff': const EdiTeXCommand(
      'diff', r'\dfrac{\partial}{\partial $0}$1~', r'diff($0,$1)'),
  'lim': const EdiTeXCommand(
      'lim', r'~\lim\limits_{$0\to $1}$2~', r'lim($0,$1,$2)'),
  'int': const EdiTeXCommand(
      'int', r'~\displaystyle\int_{$0}^{$1}$2~\text{d}$3~', r'int($3,$0,$1,$2)')
};

const _defaultInstantCommands = const {
  '(': const EdiTeXCommand('(', r'\left($0\right)', r'($0)', true),
  '[': const EdiTeXCommand('(', r'\left[$0\right]', r'brackets($0)', true),
  '^': const EdiTeXCommand('^', r'^{$0}', r'^($0)'),
  '_': const EdiTeXCommand('_', r'_{$0}', r'_($0)'),
  '*': const EdiTeXCommand('*', r'\cdot{}', '*'),
  '+': const EdiTeXCommand('+', r'+{}', '+'),
  '-': const EdiTeXCommand('-', r'-{}', '-'),
  '!': const EdiTeXCommand('!', r'!', '!')
};

const _subSupCommand =
    const EdiTeXCommand('subsup', r'_{$0}^{$1}', r'_($0)^($1)');

EdiTeXCommand _defaultCommandResolver(String command) {
  return _defaultCommands[command];
}

EdiTeXCommand _defaultInstantCommandResolver(String command) {
  return _defaultInstantCommands[command];
}

/// Wrapper of all interface functions.
class EdiTeXInterface {
  final CommandResolver resolveCommand;
  final InstantCommandResolver resolveInstantCommand;

  const EdiTeXInterface(
      {this.resolveCommand = _defaultCommandResolver,
      this.resolveInstantCommand = _defaultInstantCommandResolver});

  bool hasCommand(String command) {
    final template = resolveCommand(command);
    return template != null;
  }

  bool hasInstantCommand(String command) {
    final template = resolveInstantCommand(command);
    return template != null;
  }
}

/// Expression fragment.
abstract class Fragment {
  Fragment clone();

  String render(W<int> cursorIdx);

  bool insert(W<int> cursorIdx, Fragment fragment, EdiTeXInterface i) => false;

  bool remove(W<int> cursorIdx) => false;

  int finish(W<int> cursorIdx, EdiTeXInterface i) => 0;

  String getParsableContent() => '';

  int get length;
}

/// Fragment containing a single text character.
class CharFragment extends Fragment {
  static const overrideRender = const {r'\': r'\backslash'};

  String char;

  CharFragment(this.char);

  @override
  CharFragment clone() => new CharFragment(char);

  @override
  String render(cursorIdx) {
    if (overrideRender.containsKey(char)) {
      return overrideRender[char];
    }
    return '{$char}';
  }

  @override
  String getParsableContent() => char;

  @override
  int get length => 1;
}

/// Fragment containing a command.
class CommandFragment extends Fragment {
  static final _argRegex = new RegExp(r'\$([0-9]+)');

  final EdiTeXCommand data;
  List<MultiFragment> args;

  CommandFragment(this.data, [this.args]) {
    if (args == null) {
      args = new List<MultiFragment>();

      // Create fragment for each argument.
      data.latexTemplate.replaceAllMapped(_argRegex, (match) {
        args.add(new MultiFragment());
        return '';
      });
    }
  }

  @override
  CommandFragment clone() =>
      new CommandFragment(data, args.map((arg) => arg.clone()).toList());

  @override
  bool insert(cursorIdx, fragment, interface) {
    for (final arg in args) {
      if (arg.insert(cursorIdx, fragment, interface)) {
        return true;
      }
      cursorIdx.v--;
    }
    return false;
  }

  @override
  bool remove(cursorIdx) {
    for (final arg in args) {
      if (arg.remove(cursorIdx)) {
        return true;
      }
      cursorIdx.v--;
    }
    return false;
  }

  @override
  int finish(cursorIdx, commands) {
    for (final arg in args) {
      final result = arg.finish(cursorIdx, commands);
      if (result != 0) {
        return result;
      }
      cursorIdx.v--;
    }
    return 0;
  }

  @override
  String render(cursorIdx) {
    final tex = data.latexTemplate.replaceAllMapped(_argRegex, (match) {
      cursorIdx.v--;
      final idx = int.parse(match.group(1));
      return args[idx].render(cursorIdx);
    });
    return data.changesHeight ? '{$tex}' : tex;
  }

  @override
  String getParsableContent() {
    final str = data.parseTemplate.replaceAllMapped(_argRegex, (match) {
      final idx = int.parse(match.group(1));
      return args[idx].getParsableContent();
    });

    // Add space before and after command to make implicit multiplication
    // obvious.
    return ' $str ';
  }

  @override
  int get length => args.fold(1, (prev, arg) => prev + arg.length);
}

/// Fragment containing multiple child fragments.
class MultiFragment extends Fragment {
  List<Fragment> fragments;

  MultiFragment([this.fragments]) {
    if (fragments == null) {
      fragments = new List<Fragment>();
    }
  }

  @override
  MultiFragment clone() =>
      new MultiFragment(fragments.map((frag) => frag.clone()).toList());

  /// Sequential sub and sup commands should be replaced by a composite command.
  void _checkSequentialSubSup() {
    if (fragments.isEmpty) {
      return;
    }

    var prev = fragments.first;
    for (var i = 1; i < fragments.length; i++) {
      final curr = fragments[i];
      if (prev is CommandFragment && curr is CommandFragment) {
        MultiFragment subArg, supArg;
        if (prev.data.name == '_' && curr.data.name == '^') {
          subArg = prev.args.first;
          supArg = curr.args.first;
        } else if (curr.data.name == '_' && prev.data.name == '^') {
          subArg = curr.args.first;
          supArg = prev.args.first;
        }

        if (subArg != null && supArg != null) {
          final command = new CommandFragment(_subSupCommand, [subArg, supArg]);
          fragments.removeRange(i - 1, i + 1);
          fragments.insert(i - 1, command);
          i--;
        }
      }

      prev = curr;
    }
  }

  @override
  bool insert(cursorIdx, fragment, interface) {
    for (var i = 0; i <= fragments.length; i++) {
      if (cursorIdx.v == 0) {
        // Special case: forward slash for division, turn into frac command.
        if (fragment is CharFragment && fragment.char == '/') {
          // Get fraction numerator part.
          i--; // Go to previous fragment.
          final startI = i;
          for (; i >= 0; i--) {
            final fragment = fragments[i];
            if (fragment is CommandFragment &&
                (fragment.data.name == '+' || fragment.data.name == '-')) {
              break;
            }
          }

          i++; // -1 => 0 or +/- text fragment => first fragment of numerator.

          final numerator = new MultiFragment();
          numerator.fragments.addAll(fragments.getRange(i, startI + 1));
          fragments.removeRange(i, startI + 1);

          final template = interface.resolveCommand(_divideCommand);
          if (template != null) {
            final command = new CommandFragment(template);
            command.args[0] = numerator;
            fragments.insert(i, command);
          }
          return true;
        } else {
          fragments.insert(i, fragment);
          _checkSequentialSubSup();
          return true;
        }
      } else if (i < fragments.length) {
        cursorIdx.v--;
        if (fragments[i].insert(cursorIdx, fragment, interface)) {
          return true;
        }
      }
    }
    return false;
  }

  @override
  bool remove(cursorIdx) {
    for (var i = 0; i < fragments.length; i++) {
      if (cursorIdx.v == 0) {
        fragments.removeAt(i);
        _checkSequentialSubSup();
        return true;
      } else {
        cursorIdx.v--;
        if (fragments[i].remove(cursorIdx)) {
          return true;
        }
      }
    }
    return false;
  }

  @override
  int finish(cursorIdx, interface) {
    for (var i = 0; i < fragments.length; i++) {
      cursorIdx.v--;
      if (cursorIdx.v == 0) {
        // Append all text fragments before the cursor.
        final command = new List<int>();
        final startIdx = i;
        for (; i >= 0; i--) {
          final fragi = fragments[i];
          if (fragi is CharFragment) {
            if (fragi.char == r'\') {
              i--; // ensure removal of backslash.
              break;
            }
            command.insert(0, fragi.char.codeUnitAt(0));
          } else {
            break;
          }
        }

        final commandStr = new String.fromCharCodes(command);
        final template = interface.resolveCommand(commandStr);
        if (template != null) {
          fragments.removeRange(i + 1, startIdx + 1);
          final fragment = new CommandFragment(template);
          fragments.insert(i + 1, fragment);
          return -(startIdx - i);
        } else {
          return 1;
        }
      } else {
        final result = fragments[i].finish(cursorIdx, interface);
        if (result != 0) {
          return result;
        }
      }
    }
    return 0;
  }

  @override
  String render(cursorIdx) {
    final buffer = new StringBuffer();
    if (cursorIdx.v == 0) {
      buffer.write(_cursorCommand);
    }
    if (fragments.isNotEmpty) {
      for (var i = 0; i < fragments.length; i++) {
        final fragment = fragments[i];
        final fragTex = fragment.render(cursorIdx);
        cursorIdx.v--;

        if (cursorIdx.v == 0) {
          if (fragment is CommandFragment && fragment.data.changesHeight) {
            buffer.write('{$fragTex$_cursorCommand}');
          } else {
            buffer.write('$fragTex$_cursorCommand');
          }
        } else {
          buffer.write(fragTex);
        }
      }
    } else {
      buffer.write(_emptyCommand);
    }
    return buffer.toString();
  }

  List getData() {
    final list = new List();

    // Concat sequential char fragments.
    var prevIsChar = false;
    for (final fragment in fragments) {
      if (fragment is CharFragment) {
        if (prevIsChar) {
          // Concatenate
          list[list.length - 1] = '${list.last}${fragment.char}';
        } else {
          list.add(fragment.char);
          prevIsChar = true;
        }
      } else if (fragment is CommandFragment) {
        final commandData = new List()
          ..add(fragment.data.name)
          ..addAll(fragment.args.map((arg) => arg.getData()));
        list.add(commandData);
        prevIsChar = false;
      } else {
        throw new StateError(
            'MultiFragment contains fragments of type ${fragment.runtimeType}');
      }
    }

    return list;
  }

  void loadData(List data, EdiTeXInterface interface) {
    for (final item in data) {
      if (item is String) {
        // Unpack into multpile CharFragments.
        for (final char in item.runes) {
          fragments.add(new CharFragment(new String.fromCharCode(char)));
        }
      } else if (item is List) {
        // Construct [CommandData].
        final command = item.first;
        var commandData = interface.resolveInstantCommand(command);
        if (commandData == null) {
          commandData = interface.resolveCommand(command);
          if (commandData == null) {
            throw new ArgumentError("command '$command' cannot be resolved");
          }
        }

        // Construct arguments.
        final args = new List<MultiFragment>();
        for (var i = 1; i < item.length; i++) {
          args.add(new MultiFragment()..loadData(item[i], interface));
        }

        fragments.add(new CommandFragment(commandData, args));
      }
    }
  }

  @override
  String getParsableContent() =>
      fragments.fold('', (prev, frag) => '$prev${frag.getParsableContent()}');

  @override
  int get length => fragments.fold(1, (prev, arg) => prev + arg.length);
}

/// For [EdiTeX.undoHistory] and [EdiTeX.redoHistory].
class _EditHistoryEntry {
  final Fragment content;
  final int cursorIdx;
  _EditHistoryEntry(this.content, this.cursorIdx);
}

/// EdiTeX editor component.
/// Can initialize itself in an empty <div>
class EdiTeX {
  final EdiTeXInterface interface;
  final DivElement container, katexWrapper;

  final undoHistory = new List<_EditHistoryEntry>();
  final redoHistory = new List<_EditHistoryEntry>();
  var content = new MultiFragment();

  var cursorIdx = 0;
  var doUpdate = true;

  final _onLeftLeave = new StreamController<Null>.broadcast();
  final _onRightLeave = new StreamController<Null>.broadcast();

  EdiTeX(this.container, [this.interface = const EdiTeXInterface()])
      : katexWrapper = new DivElement() {
    container.append(katexWrapper);

    // Make focussable via keyboard.
    container.tabIndex = 0;

    // Add keyboard listeners.
    container.onKeyDown.listen(onKeyDown);

    new Timer.periodic(new Duration(milliseconds: 30), (t) {
      if (doUpdate) {
        // Set [doUpdate] to false first in case it is updated again while
        // rendering.
        doUpdate = false;

        // Render content using KaTeX.
        katex.render(getLaTeX(), katexWrapper,
            new katex.RenderingOptions(displayMode: true));

        // Find cursor.
        final cursor = katexWrapper
            .querySelector("span.mspace[style='color: transparent;']");
        cursor.classes.add('editex-cursor');
        final katexDisplay = katexWrapper.children.first;

        // Compute katexDiv offset using cursor position
        var cursorOffset = cursor.getBoundingClientRect().left;
        final wrapperWidth = katexWrapper.getBoundingClientRect().width;
        final katexRect = katexDisplay.getBoundingClientRect();
        cursorOffset -= katexRect.left - 1; // -1 because of cursor margin.

        if (cursorOffset > wrapperWidth) {
          katexDisplay.style.marginLeft = '-${cursorOffset - wrapperWidth}px';
        } else {
          katexDisplay.style.marginLeft = '0px';
        }
      }
    });
  }

  Future destroy() {
    return Future.wait([_onLeftLeave.close(), _onRightLeave.close()]);
  }

  Stream<Null> get onLeftLeave => _onLeftLeave.stream;
  Stream<Null> get onRightLeave => _onRightLeave.stream;

  /// Add undo checkpoint.
  void addUndoCheckpoint() {
    undoHistory.add(new _EditHistoryEntry(content.clone(), cursorIdx));
    redoHistory.clear();
  }

  /// Undo last change.
  void undo() {
    if (undoHistory.isNotEmpty) {
      redoHistory.add(new _EditHistoryEntry(content.clone(), cursorIdx));
      final replace = undoHistory.removeLast();
      content = replace.content;
      cursorIdx = replace.cursorIdx;
      doUpdate = true;
    }
  }

  /// Redo last change.
  void redo() {
    if (redoHistory.isNotEmpty) {
      undoHistory.add(new _EditHistoryEntry(content.clone(), cursorIdx));
      final replace = redoHistory.removeLast();
      content = replace.content;
      cursorIdx = replace.cursorIdx;
      doUpdate = true;
    }
  }

  /// Keyboard handler.
  void onKeyDown(KeyboardEvent event) {
    if (event.ctrlKey) {
      // Keystrokes
      if (event.keyCode == KeyCode.Z) {
        if (event.shiftKey) {
          redo();
        } else {
          undo();
        }
      }
    } else if (new RegExp(r'^[0-9A-Za-z\\]$').hasMatch(event.key)) {
      addUndoCheckpoint();

      var letter = event.key;
      if (event.shiftKey) {
        letter = letter.toUpperCase();
      }

      // Insert letter into content.
      content.insert(
          new W<int>(cursorIdx), new CharFragment(letter), interface);
      cursorIdx++;
    } else if (interface.hasInstantCommand(event.key)) {
      addUndoCheckpoint();

      final fragment =
          new CommandFragment(interface.resolveInstantCommand(event.key));
      content.insert(new W<int>(cursorIdx), fragment, interface);
      cursorIdx++;
    } else if (event.key == '/') {
      // Only proceed if [_divideCommand] is specified.
      if (interface.hasCommand(_divideCommand)) {
        addUndoCheckpoint();

        content.insert(new W<int>(cursorIdx), new CharFragment('/'), interface);
        cursorIdx += 2;
      }
    } else if (event.keyCode == KeyCode.SPACE) {
      addUndoCheckpoint();

      final result = content.finish(new W<int>(cursorIdx), interface);
      if (result < 0) {
        cursorIdx += result + 1;
      }
    } else if (event.keyCode == KeyCode.BACKSPACE) {
      addUndoCheckpoint();

      if (cursorIdx > 0) {
        content.remove(new W<int>(cursorIdx - 1));
        cursorIdx--;
      }
    } else if (event.keyCode == KeyCode.DELETE) {
      content.remove(new W<int>(cursorIdx));
    } else if (event.keyCode == KeyCode.LEFT) {
      cursorIdx--;
      if (cursorIdx == -1) {
        cursorIdx = 0;
        _onLeftLeave.add(null);
        return;
      }
    } else if (event.keyCode == KeyCode.RIGHT) {
      cursorIdx++;
      if (cursorIdx == content.length) {
        cursorIdx--;
        _onRightLeave.add(null);
        return;
      }
    } else if (event.keyCode == KeyCode.HOME) {
      cursorIdx = 0;
    } else if (event.keyCode == KeyCode.END) {
      cursorIdx = content.length - 1;
    } else {
      // Prevent update.
      return;
    }

    doUpdate = true;
  }

  /// Get JSON serializable data object.
  List getData() => content.getData();

  /// Load data generated with [getData].
  void loadData(List data) {
    addUndoCheckpoint();
    content = new MultiFragment()..loadData(data, interface);
    cursorIdx = 0;
    doUpdate = true;
  }

  /// Get content as parsable expression.
  String getParsableContent() => content.getParsableContent().trim();

  /// Get LaTeX string with cursor.
  String getLaTeX() => content.render(new W<int>(cursorIdx));
}

class W<T> {
  T v;
  W(this.v);
}
